/*****************************************************************************
  idr.cu

  (c) 2014 - Nikhil R Podduturi, J. Seth Strattan
  J. Michael Cherry Lab, Department of Genetics, Stanford University School of Medicine

  Licensed under the GNU General Public License 2.0 license.
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include <ostream>
#include <fstream>
#include <nvbio/basic/console.h>
#include <nvbio/basic/timer.h>

#include "ranker.h"
#include "processPeaks.h"
#include "idr_kernel.h"

void crcInit();

using namespace nvbio;

// Utility struct to keep track of final results
struct overlap
{
    std::string chr;
    uint32 start1;
    uint32 end1;
    double rankingMeasure1;
    uint32 start2;
    uint32 end2;
    double rankingMeasure2;
    double idrLocal;
    double idr;
};

int main(int argc, char* argv[])
{

    Timer timer;
    timer.start();

    cudaSetDeviceFlags( cudaDeviceMapHost | cudaDeviceLmemResizeToMax );

    crcInit();

    if (argc == 1 ||
            (argc == 2 && strcmp( argv[1], "--help" ) == 0) ||
            (argc == 2 && strcmp( argv[1], "-h" ) == 0))
    {
        log_info(stderr,"idr [options] --peak-files peak1 peak2 --genome-file genomefile\n");
        log_info(stderr,"Required Params:\n");
        log_info(stderr,"       --peak-files            Set of narrowPeak or broadPeak files\n");
        log_info(stderr,"       --genome-file           Genome file with chromosomes size\n");
        log_info(stderr,"options:\n");
        log_info(stderr,"   General:\n");
        log_info(stderr,"       --idr-cutoff            IDR cutoff (default: 0.025)\n");
        log_info(stderr,"       --target-type           Type of assay target (default: tf)\n");
        log_info(stderr,"       --ranking-measure       Type of ranking measure (default: signal.value)\n");
        log_info(stderr,"Usage:\n");
        log_info(stderr,"   For SPP:\n");
        log_info(stderr,"       idr --ranking-measure signal.value --peak-files peak1 peak2 --genome-file genome_table.txt\n");
        log_info(stderr,"   For MACSV2:\n");
        log_info(stderr,"       idr --ranking-measure p.value --peak-files peak1 peak2 --genome-file genome_table.txt\n");
        exit(0);
    }

    // Default values for parameters
    std::string rankingMeasure    =   "signalValue";
    std::string target_type =   "tf";
    std::string peakFileA   =   "";
    std::string peakFileB   =   "";
    std::string genomeFile  =   "";
    float idrCutoff = 0.025;

    // Handle user params and options
    for (int32 i = 1; i < argc; ++i)
    {
        if (strcmp( argv[i], "-ranking-measure" ) == 0 ||
            strcmp( argv[i], "--ranking-measure" ) == 0)
        {
            if (strcmp( argv[i+1], "p.value" ) == 0)
                rankingMeasure = "pValue";
            else if (strcmp( argv[i+1], "q.value" ) == 0)
                rankingMeasure = "qValue";
        }
        else if (strcmp( argv[i], "-peak-files") == 0 ||
                strcmp( argv[i], "--peak-files") == 0)
        {
            peakFileA   =   argv[i + 1];
            peakFileB   =   argv[i + 2];
        }
        else if (strcmp( argv[i], "-genome-file") == 0 ||
                strcmp( argv[i], "--genome-file") == 0)
        {
            genomeFile  =   argv[i + 1];
        }
        else if (strcmp( argv[i], "-idr-cutoff") == 0 ||
                strcmp( argv[i], "--idr-cutoff") == 0)
        {
            idrCutoff = atof(argv[i + 1]);
        }
    }
    // inspect and select cuda devices
    int cuda_device  = -1;
    int device_count;
    cudaGetDeviceCount(&device_count);
    log_verbose(stderr, "  cuda devices : %d\n", device_count);

    if (device_count)
    {
        if (cuda_device == -1)
        {
            int best_device = 0;
            cudaDeviceProp best_device_prop;
            cudaGetDeviceProperties( &best_device_prop, best_device );
            for (int device = 0; device < device_count; ++device)
            {
                cudaDeviceProp device_prop;
                cudaGetDeviceProperties( &device_prop, device );
                if (device_prop.major >= best_device_prop.major &&
                        device_prop.minor >= best_device_prop.minor)
                {
                    best_device_prop = device_prop;
                    best_device      = device;
                }
            }
            cuda_device = best_device;
        }
        log_verbose(stderr, "  chosen device %d\n", cuda_device);
        {
            cudaDeviceProp device_prop;
            cudaGetDeviceProperties( &device_prop, cuda_device );
            log_verbose(stderr, "    device name        : %s\n", device_prop.name);
            log_verbose(stderr, "    compute capability : %d.%d\n", device_prop.major, device_prop.minor);
        }
        cudaSetDevice( cuda_device );
    }

    // Processing peaks here using bedtools and then calculate overlaps
    ProcessPeaks *bc = new ProcessPeaks(peakFileA, peakFileB, genomeFile);

    // Merged vectors to run IDR
    std::vector<double> merge_A, merge_B, unmatched_merge_A, unmatched_merge_B;
    std::vector<unsigned int> tracker;
    std::vector<overlap> overlaps;

    unsigned int start = 0;
    for (size_t i = 0; i < bc->_peakA->bedList.size(); ++i)
    {
        if (rankingMeasure == "signalValue")
            merge_A.push_back(bc->_peakA->bedList[i].signalValue);
        else if(rankingMeasure == "pValue")
            merge_A.push_back(bc->_peakA->bedList[i].pValue);
        else if(rankingMeasure == "qValue")
            merge_A.push_back(bc->_peakA->bedList[i].qValue);

        merge_B.push_back(0);
        while (start < bc->overlap_index_A[i])
        {

            tracker.push_back( bc->overlap_index_B[start] );
            double bSigVal = 0.0;

            if (rankingMeasure == "signalValue")
                bSigVal = bc->_peakB->bedList[ bc->overlap_index_B[start] ].signalValue;
            else if (rankingMeasure == "pValue")
                bSigVal = bc->_peakB->bedList[ bc->overlap_index_B[start] ].pValue;
            else if (rankingMeasure == "qValue")
                bSigVal = bc->_peakB->bedList[ bc->overlap_index_B[start] ].qValue;

            overlap o;
            std::string chr(bc->_peakA->bedList[i].chrom);
            o.chr = chr;
            o.start1 = bc->_peakA->bedList[i].start;
            o.end1 = bc->_peakA->bedList[i].end;
            o.rankingMeasure1 = merge_A[i];
            o.start2 = bc->_peakB->bedList[ bc->overlap_index_B[start] ].start;
            o.end2 = bc->_peakB->bedList[ bc->overlap_index_B[start] ].end;

            // Averging the ranking measure if one peak overlaps with multiple peaks
            if (merge_B[i] == 0)
            {
                merge_B[i] = bSigVal;
                o.rankingMeasure2 = merge_B[i];
                overlaps.push_back(o);
            }
            else
            {
                merge_B[i] = double (merge_B[i] + bSigVal)/2.0;
                o.rankingMeasure2 = merge_B[i];
                overlaps.pop_back();
                overlaps.push_back(o);
            }
            ++start;
        }
    }

    std::sort(tracker.begin(), tracker.end());
    for (unsigned int i = 0; i < bc->_peakB->bedList.size(); ++i)
    {
        if( std::find(tracker.begin(), tracker.end(), i) == tracker.end() )
        {
            double bSigVal = 0.0;
            if (rankingMeasure == "signalValue")
                bSigVal = bc->_peakB->bedList[i].signalValue;
            else if (rankingMeasure == "pValue")
                bSigVal = bc->_peakB->bedList[i].pValue;
            else if (rankingMeasure == "qValue")
                bSigVal = bc->_peakB->bedList[i].qValue;

            merge_A.push_back( 0 );
            merge_B.push_back( bSigVal );
        }
    }

    for (unsigned int i=0; i < merge_A.size(); ++i)
    {
        if ( merge_A[i] != 0 && merge_B[i] != 0 )
        {
            unmatched_merge_A.push_back( -merge_A[i] );
            unmatched_merge_B.push_back( -merge_B[i] );
        }
    }

    std::vector<float> ranks_A, ranks_B;

    rank( unmatched_merge_A, ranks_A, "average" );
    rank( unmatched_merge_B, ranks_B, "average" );

    log_info(stderr, "Number of overlaps after removing duplicates - %d\n", ranks_A.size());

    log_info(stderr, "Fit 2-component model - started\n");
    thrust::device_vector<double> idrLocal( ranks_A.size() );
    em_gaussian(ranks_A, ranks_B, idrLocal);

    thrust::device_vector<double> positions (idrLocal.size());
    thrust::sequence(positions.begin(), positions.end(), 1.0, 1.0);

    thrust::device_vector<double> temp = positions;

    thrust::device_vector<double> idr = idrLocal;
    thrust::sort_by_key(idr.begin(), idr.end(), positions.begin());

    thrust::inclusive_scan(thrust::device, idr.begin(), idr.end(), idr.begin());
    thrust::transform(thrust::device, idr.begin(), idr.end(),
            temp.begin(), idr.begin(), thrust::divides<double>());

    temp.clear();
    using namespace thrust::placeholders;
    int counter = thrust::count_if(idr.begin(), idr.end(), _1 <= idrCutoff);
    thrust::sort_by_key(positions.begin(), positions.end(), idr.begin());

    std::filebuf fb;
    fb.open ("idrValues.txt",std::ios::out);
    std::ostream fout(&fb);
    fout.precision(15);
    for (unsigned int i=0; i < idr.size(); ++i)
    {
        overlaps[i].idrLocal = idrLocal[i];
        overlaps[i].idr = idr[i];
        if (overlaps[i].idr <= idrCutoff)
        {
            fout<<overlaps[i].chr<<"\t"<<
                overlaps[i].start1<<"\t"<<
                overlaps[i].end1<<"\t"<<
                overlaps[i].rankingMeasure1<<"\t"<<
                overlaps[i].start2<<"\t"<<
                overlaps[i].end2<<"\t"<<
                overlaps[i].rankingMeasure2<<"\t"<<
                std::fixed<<overlaps[i].idrLocal<<"\t"<<
                std::fixed<<overlaps[i].idr<<endl;
        }
    }
    fb.close();
    log_info(stderr, "Number of peaks passing IDR cutoff of %f - %d\n", idrCutoff, counter);

    timer.stop();
    log_info(stderr, "Time taken - %um:%us\n",
        uint32(timer.seconds()/60),
        uint32(timer.seconds())%60);

    return 0;
}
