/*********************************************************************
   idr_kernel.h

   (c) 2014 - Nikhil R Podduturi, J. Seth Strattan
   J. Michael Cherry Lab, Department of Genetics, Stanford University School of Medicine

   Licensed under the GNU General Public License 2.0 license.
**********************************************************************/

#include <float.h>
#include <iomanip>

#include <thrust/sort.h>
#include <thrust/scan.h>
#include <thrust/find.h>
#include <thrust/fill.h>
#include <thrust/count.h>
#include <thrust/replace.h>
#include <thrust/transform.h>
#include <thrust/sequence.h>
#include <thrust/device_vector.h>
#include <thrust/functional.h>
#include <thrust/execution_policy.h>
#include <thrust/binary_search.h>


/* Utility functions for error handling and printing vectors from device */
#define gpuErrchk(ans) { gpuAssert((ans), __FILE__, __LINE__); }
inline void gpuAssert(cudaError_t code, char *file, int line, bool abort=true)
{
    if (code != cudaSuccess)
    {
        fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
        if (abort) exit(code);
    }
}

template <typename Vector>
void print_vector(const std::string& name, const Vector& v)
{
    typedef typename Vector::value_type T;
    std::cout << "  " << std::setw(20) << name << "  ";
    thrust::copy(v.begin(), v.end(), std::ostream_iterator<T>(std::cout, " "));
    std::cout << std::endl;
}


/* IDR GPU kernels */
__global__ void calculate_cdf(float* input, double* cdf, double* pdf,
        double* temp_cdf, double* breaks, int n)
{
    int tid = threadIdx.x+blockIdx.x*blockDim.x;
    if( tid < n )
    {
        int index = lroundf(cdf[tid]);
        double td = temp_cdf[index-1];
        double pf =  pdf[index-1];
        cdf[tid] = td + pf*(input[tid]-breaks[index-1]);
    }
}

__global__ void calculate_quantiles(double* x_cdf,
        double* y_cdf, double rho,
        double* density, int n)
{
    int tid = threadIdx.x+blockIdx.x*blockDim.x;
    if ( tid < n)
    {
        double a = pow(normcdfinv(x_cdf[tid]), 2) + pow(normcdfinv(y_cdf[tid]), 2);
        double b = normcdfinv(x_cdf[tid]) * normcdfinv(y_cdf[tid]);
        density[tid] = exp( -log(1 - pow(rho, 2)) / 2 - rho / (2 * (1 - pow(rho, 2))) * (rho*a-2*b) );
    }
}

__global__ void multiply_ez(double* density, double* d, double *ez, int n)
{
    int tid = threadIdx.x+blockIdx.x*blockDim.x;
    if ( tid < n) {
        d[tid] = ez[tid] * log(density[tid]);
    }
}

__global__ void sum_likelihood(double* likelihood, double* density_1, double* x1_pdf,
    double* y1_pdf, double* x2_pdf, double* y2_pdf, double p, int n)
{
    int tid = threadIdx.x+blockIdx.x*blockDim.x;
    if ( tid < n) {
        likelihood[tid] = log(p * density_1[tid] * x1_pdf[tid] * y1_pdf[tid] + (1.0 - p) * 1.0 * x2_pdf[tid] * y2_pdf[tid]);
    }
}

__global__ void get_ez(double* ez, double* density_1, double* x1_pdf,
        double* y1_pdf, double* x2_pdf, double* y2_pdf, double p, int n)
{
    int idx = threadIdx.x + blockDim.x*blockIdx.x;
    if (idx < n)
    {
        ez[idx] = p*density_1[idx]*x1_pdf[idx]*y1_pdf[idx]/(p * density_1[idx] * x1_pdf[idx] * y1_pdf[idx] + (1-p) * 1 * x2_pdf[idx] * y2_pdf[idx]);
    }
}

template <typename T>
__global__ void find(T *data, T value, double* ez, double *sum, double* sum_1, int n)
{
    int idx = threadIdx.x + blockDim.x*blockIdx.x;
    if ( idx < n ){
        if(data[idx] == value){
            sum[idx] = ez[idx];
            sum_1[idx] = 1 - ez[idx];
        }
    }
} // end of GPU kernels

double cost_function(double* x_cdf_ptr, double* y_cdf_ptr,
    double* ez_ptr, double rho, int n)
{
    thrust::device_vector<double> density_vec(n);
    thrust::device_vector<double> new_density(n);

    double* density_ptr = thrust::raw_pointer_cast( &density_vec[0] );
    double* new_density_ptr = thrust::raw_pointer_cast( &new_density[0] );

    int gridDim = round( (float) n/256 ) + 1;

    calculate_quantiles<<<gridDim, 256, 1>>>(x_cdf_ptr, y_cdf_ptr,
            rho, density_ptr, n);
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    multiply_ez<<<gridDim, 256, 1>>>(density_ptr, new_density_ptr, ez_ptr, n);
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    double cop_den = thrust::reduce(new_density.begin(), new_density.end());
    return -cop_den;
}

template <typename V>
double maximum_likelihood(V& x_cdf, V& y_cdf, V& ez)
{
    typedef typename V::value_type ValueType;

    ValueType* x_cdf_ptr = thrust::raw_pointer_cast( &x_cdf[0] );
    ValueType* y_cdf_ptr = thrust::raw_pointer_cast( &y_cdf[0] );
    ValueType* ez_ptr = thrust::raw_pointer_cast( &ez[0] );

    double ax = -0.998;
    double bx = 0.998;
    double tol = 0.00001;

    /*  c is the squared inverse of the golden ratio */
    const double c = (3. - sqrt(5.)) * .5;

    /* Local variables */
    double a, b, d, e, p, q, r, u, v, w, x;
    double t2, fu, fv, fw, fx, xm, eps, tol1, tol3;

    /*  eps is approximately the square root of the relative machine precision. */
    eps = DBL_EPSILON;
    tol1 = eps + 1.;/* the smallest 1.000... > 1 */
    eps = sqrt(eps);

    a = ax;
    b = bx;
    v = a + c * (b - a);
    w = v;
    x = v;

    d = 0.;/* -Wall */
    e = 0.;
    fx = cost_function(x_cdf_ptr, y_cdf_ptr, ez_ptr, x, x_cdf.size());
    fv = fx;
    fw = fx;
    tol3 = tol / 3.;

    for(;;) {
        xm = (a + b) * .5;
        tol1 = eps * fabs(x) + tol3;
        t2 = tol1 * 2.;

        /* check stopping criterion */
        if (fabs(x - xm) <= t2 - (b - a) * .5) break;
        p = 0.;
        q = 0.;
        r = 0.;
        if (fabs(e) > tol1) { /* fit parabola */
            r = (x - w) * (fx - fv);
            q = (x - v) * (fx - fw);
            p = (x - v) * q - (x - w) * r;
            q = (q - r) * 2.;
            if (q > 0.) p = -p; else q = -q;
            r = e;
            e = d;
        }

        if (fabs(p) >= fabs(q * .5 * r) ||
            p <= q * (a - x) || p >= q * (b - x)) { /* a golden-section step */
            if (x < xm) e = b - x; else e = a - x;
            d = c * e;
        }
        else { /* a parabolic-interpolation step */
            d = p / q;
            u = x + d;
            /* f must not be evaluated too close to ax or bx */
            if (u - a < t2 || b - u < t2) {
            d = tol1;
            if (x >= xm) d = -d;
            }
        }

        /* f must not be evaluated too close to x */
        if (fabs(d) >= tol1)
            u = x + d;
        else if (d > 0.)
            u = x + tol1;
        else
            u = x - tol1;

        fu = cost_function(x_cdf_ptr, y_cdf_ptr, ez_ptr, u, x_cdf.size());

        /*  update  a, b, v, w, and x */
        if (fu <= fx) {
            if (u < x) b = x; else a = x;
            v = w;    w = x;   x = u;
            fv = fw; fw = fx; fx = fu;
        } else {
            if (u < x) a = u; else b = u;
            if (fu <= fw || w == x) {
            v = w; fv = fw;
            w = u; fw = fu;
            } else if (fu <= fv || v == x || v == w) {
            v = u; fv = fu;
            }
        }
    }
    return x;
}

template <typename V>
double gaussian_loglikelihood(V& x1_pdf, V& x2_pdf, V& x1_cdf, V& x2_cdf,
        V& y1_pdf, V& y2_pdf, V& y1_cdf, V& y2_cdf, double p, double rho)
{
    typedef typename V::value_type ValueType;

    int n = x1_pdf.size();
    int gridDim = round( (float) x1_pdf.size()/256 ) + 1;

    thrust::device_vector<ValueType> density_c1( n );
    thrust::device_vector<ValueType> likelihood_vec( n );

    ValueType* density_c1_ptr = thrust::raw_pointer_cast( &density_c1[0] );
    ValueType* likelihood_ptr = thrust::raw_pointer_cast( &likelihood_vec[0] );

    ValueType* x1_pdf_ptr = thrust::raw_pointer_cast( &x1_pdf[0] );
    ValueType* x2_pdf_ptr = thrust::raw_pointer_cast( &x2_pdf[0] );
    ValueType* x1_cdf_ptr = thrust::raw_pointer_cast( &x1_cdf[0] );
    ValueType* x2_cdf_ptr = thrust::raw_pointer_cast( &x2_cdf[0] );
    ValueType* y1_pdf_ptr = thrust::raw_pointer_cast( &y1_pdf[0] );
    ValueType* y2_pdf_ptr = thrust::raw_pointer_cast( &y2_pdf[0] );
    ValueType* y1_cdf_ptr = thrust::raw_pointer_cast( &y1_cdf[0] );
    ValueType* y2_cdf_ptr = thrust::raw_pointer_cast( &y2_cdf[0] );

    calculate_quantiles<<<gridDim, 256, 1>>>(x1_cdf_ptr, y1_cdf_ptr,
        rho, density_c1_ptr, n);
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    sum_likelihood<<<gridDim, 256, 1>>>(likelihood_ptr, density_c1_ptr,
        x1_pdf_ptr, y1_pdf_ptr, x2_pdf_ptr, y2_pdf_ptr, p, n);
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    double l0 = thrust::reduce(likelihood_vec.begin(), likelihood_vec.end());

    density_c1.clear();
    density_c1.shrink_to_fit();

    likelihood_vec.clear();
    likelihood_vec.shrink_to_fit();

    return l0;
}

template <typename V, typename T>
void estimate_marginals(V& input, T& breaks, T& pdf_1, T& pdf_2,
        T& cdf_1, T& cdf_2, T& ez, double p)
{
    int nbins = breaks.size() - 1;
    int input_size = input.size();

    thrust::host_vector<double> temp_cdf_1( nbins ), temp_cdf_2( nbins );
    thrust::host_vector<double> temp_pdf_1( nbins ), temp_pdf_2( nbins );

    thrust::lower_bound(breaks.begin(), breaks.end(),
            input.begin(), input.end(),
            cdf_1.begin());

    int first_size = round((double)(input_size*p));
    double bin_width = breaks[1] - breaks[0];

    cdf_2 = cdf_1;
    pdf_1 = cdf_1;
    pdf_2 = cdf_1;

    double sum_ez = thrust::reduce(ez.begin(), ez.end());
    thrust::device_vector<double> dup_ez( ez.size() );

    using namespace thrust::placeholders;
    thrust::transform(ez.begin(), ez.end(), dup_ez.begin(), (double)1.0 - _1);

    double dup_sum_ez = thrust::reduce(dup_ez.begin(), dup_ez.end());

    int gridDim = round( (float) input.size()/256 ) + 1;
    double* cdf_1_ptr = thrust::raw_pointer_cast( &cdf_1[0] );
    double* ez_ptr = thrust::raw_pointer_cast( &ez[0] );

    for(size_t i=1; i<breaks.size(); ++i)
    {
        thrust::device_vector<double> sum_vec( input_size );
        thrust::device_vector<double> sum_vec_1( input_size );

        thrust::fill(sum_vec.begin(), sum_vec.end(), 0.0);
        double* sum_vec_ptr = thrust::raw_pointer_cast( &sum_vec[0] );

        thrust::fill(sum_vec_1.begin(), sum_vec_1.end(), 0.0);
        double* sum_vec_1_ptr = thrust::raw_pointer_cast( &sum_vec_1[0] );

        find<<<gridDim, 256, 1>>> (cdf_1_ptr, (double)i, ez_ptr, sum_vec_ptr, sum_vec_1_ptr, input_size);
        gpuErrchk( cudaPeekAtLastError() );
        gpuErrchk( cudaDeviceSynchronize() );

        double sum_1 = thrust::reduce(sum_vec.begin(), sum_vec.end());
        double sum_2 = thrust::reduce(sum_vec_1.begin(), sum_vec_1.end());

        temp_pdf_1[i-1] = (sum_1 + 1) / (sum_ez + nbins) / bin_width*(input_size+50) / (input_size+51);
        temp_pdf_2[i-1] = (sum_2 + 1) / (dup_sum_ez + nbins) / bin_width*(input_size+50) / (input_size+51);

        thrust::replace(pdf_1.begin(), pdf_1.end(), (double)i, temp_pdf_1[i-1]);
        thrust::replace(pdf_2.begin(), pdf_2.end(), (double)i, temp_pdf_2[i-1]);

        temp_cdf_1[i-1] = temp_pdf_1[i-1] * bin_width;
        temp_cdf_2[i-1] = temp_pdf_2[i-1] * bin_width;

        sum_vec.clear();
        sum_vec_1.clear();
    }

    thrust::device_vector<double> d_temp_pdf_1 = temp_pdf_1;
    thrust::device_vector<double> d_temp_cdf_1 = temp_cdf_1;
    thrust::device_vector<double> d_temp_pdf_2 = temp_pdf_2;
    thrust::device_vector<double> d_temp_cdf_2 = temp_cdf_2;

    thrust::exclusive_scan(thrust::device, d_temp_cdf_1.begin(), d_temp_cdf_1.end(), d_temp_cdf_1.begin());
    float* input_ptr = thrust::raw_pointer_cast( &input[0] );
    double* breaks_ptr = thrust::raw_pointer_cast( &breaks[0] );
    double* pdf_1_ptr = thrust::raw_pointer_cast( &d_temp_pdf_1[0] );
    double* temp_cdf_1_ptr = thrust::raw_pointer_cast( &d_temp_cdf_1[0] );
    calculate_cdf<<<gridDim, 256, 1>>>(input_ptr, cdf_1_ptr, pdf_1_ptr,
            temp_cdf_1_ptr, breaks_ptr, input.size());
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    thrust::exclusive_scan(thrust::device, d_temp_cdf_2.begin(), d_temp_cdf_2.end(), d_temp_cdf_2.begin());
    double* cdf_2_ptr = thrust::raw_pointer_cast( &cdf_2[0] );
    double* pdf_2_ptr = thrust::raw_pointer_cast( &d_temp_pdf_2[0] );
    double* temp_cdf_2_ptr = thrust::raw_pointer_cast( &d_temp_cdf_2[0] );
    calculate_cdf<<<gridDim, 256, 1>>>(input_ptr, cdf_2_ptr, pdf_2_ptr,
            temp_cdf_2_ptr, breaks_ptr, input.size());
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    /* Clearing local device vectors jus incase */
    d_temp_pdf_1.clear();
    d_temp_pdf_2.clear();
    d_temp_cdf_1.clear();
    d_temp_cdf_2.clear();

    d_temp_pdf_1.shrink_to_fit();
    d_temp_pdf_2.shrink_to_fit();
    d_temp_cdf_1.shrink_to_fit();
    d_temp_cdf_2.shrink_to_fit();
}

// TODO: duplication of code, should be refactored
template <typename V>
void estep_gaussian(V& x1_pdf, V& x2_pdf, V& x1_cdf, V& x2_cdf,
        V& y1_pdf, V& y2_pdf, V& y1_cdf, V& y2_cdf, V& ez, double p, double rho)
{
    typedef typename V::value_type ValueType;

    int n = x1_pdf.size();
    int gridDim = round( (float) x1_pdf.size()/256 ) + 1;

    thrust::device_vector<ValueType> density_c1( n );

    ValueType* density_c1_ptr = thrust::raw_pointer_cast( &density_c1[0] );
    ValueType* ez_ptr = thrust::raw_pointer_cast( &ez[0] );

    ValueType* x1_pdf_ptr = thrust::raw_pointer_cast( &x1_pdf[0] );
    ValueType* x2_pdf_ptr = thrust::raw_pointer_cast( &x2_pdf[0] );
    ValueType* x1_cdf_ptr = thrust::raw_pointer_cast( &x1_cdf[0] );
    ValueType* x2_cdf_ptr = thrust::raw_pointer_cast( &x2_cdf[0] );
    ValueType* y1_pdf_ptr = thrust::raw_pointer_cast( &y1_pdf[0] );
    ValueType* y2_pdf_ptr = thrust::raw_pointer_cast( &y2_pdf[0] );
    ValueType* y1_cdf_ptr = thrust::raw_pointer_cast( &y1_cdf[0] );
    ValueType* y2_cdf_ptr = thrust::raw_pointer_cast( &y2_cdf[0] );

    calculate_quantiles<<<gridDim, 256, 1>>>(x1_cdf_ptr, y1_cdf_ptr,
            rho, density_c1_ptr, n);
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    get_ez<<<gridDim, 256, 1>>>(ez_ptr, density_c1_ptr, x1_pdf_ptr,
            y1_pdf_ptr, x2_pdf_ptr, y2_pdf_ptr, p, n);
    gpuErrchk( cudaPeekAtLastError() );
    gpuErrchk( cudaDeviceSynchronize() );

    density_c1.clear();
    density_c1.shrink_to_fit();
}

template <typename V, typename T, typename P, typename R>
void mstep_gaussian(V& x, V& y, T& breaks, P *p0,  R *rho,
        T& x1_pdf, T& x2_pdf, T& x1_cdf, T& x2_cdf, T& y1_pdf,
        T& y2_pdf, T& y1_cdf, T& y2_cdf, T& ez)
{
    int n = x.size();

    estimate_marginals(x, breaks, x1_pdf, x2_pdf, x1_cdf, x2_cdf, ez, *p0);
    estimate_marginals(y, breaks, y1_pdf, y2_pdf, y1_cdf, y2_cdf, ez, *p0);

    *rho = maximum_likelihood(x1_cdf, y1_cdf, ez);

    double sum_ez = thrust::reduce(ez.begin(), ez.end());
    *p0 = sum_ez/(double)n;
}

template <typename V, typename T>
void em_gaussian(V& x, V& y, T& idrLocal)
{
    typedef typename V::value_type ValueType;

    thrust::device_vector<ValueType> d_ranks_A = x;
    thrust::device_vector<ValueType> d_ranks_B = y;

    thrust::device_vector<double> ez( x.size() );
    thrust::device_vector<double> breaks( 51 );

    int mid = round((float) x.size()/2);

    thrust::fill(ez.begin(), ez.begin()+mid, 0.9);
    thrust::fill(ez.begin()+mid, ez.end(), 0.1);

    float bin_width = (float)(x.size()-1)/50;

    // Constants chosen by Li and Anshul
    double p0 = 0.5;
    double rho = 0.0;
    double eps = 0.01;

    thrust::host_vector<double> likelihood;

    thrust::sequence(thrust::device, breaks.begin(),
            breaks.end(), (double)1-bin_width/100, (double)(x.size()-1+bin_width/50)/50);

    thrust::device_vector<double> x1_pdf(x.size()), x2_pdf(x.size()), x1_cdf(x.size()), x2_cdf(x.size());
    thrust::device_vector<double> y1_pdf(y.size()), y2_pdf(y.size()), y1_cdf(y.size()), y2_cdf(y.size());

    mstep_gaussian(d_ranks_A, d_ranks_B, breaks, &p0, &rho, x1_pdf, x2_pdf,
            x1_cdf, x2_cdf, y1_pdf, y2_pdf, y1_cdf, y2_cdf, ez);

    double li = gaussian_loglikelihood(x1_pdf, x2_pdf, x1_cdf, x2_cdf,
            y1_pdf, y2_pdf, y1_cdf, y2_cdf, p0, rho);

    likelihood.push_back(li);

    bool flag = true;
    int i = 1;

    //can do better. Jus replicating IDR R style coding
    int iter_counter = 1;
    while(flag)
    {
        estep_gaussian(x1_pdf, x2_pdf, x1_cdf, x2_cdf,
                 y1_pdf, y2_pdf, y1_cdf, y2_cdf, ez, p0, rho);

        mstep_gaussian(d_ranks_A, d_ranks_B, breaks, &p0, &rho, x1_pdf, x2_pdf,
                x1_cdf, x2_cdf, y1_pdf, y2_pdf, y1_cdf, y2_cdf, ez);

        double l = gaussian_loglikelihood(x1_pdf, x2_pdf, x1_cdf, x2_cdf,
                y1_pdf, y2_pdf, y1_cdf, y2_cdf, p0, rho);
        likelihood.push_back(l);

        if (i > 1)
        {
            double a_cri = likelihood[i-2] + (likelihood[i-1] - likelihood[i-2])/(1-(likelihood[i]-likelihood[i-1])/(likelihood[i-1]-likelihood[i-2]));
            if ( abs(a_cri-likelihood[i]) <= eps )
            {
                flag = false;
            }
        }
        i++;
        iter_counter++;
    }
    log_info(stderr, "Finished running IDR on the datasets\n");
    log_info(stderr, "Final P value = %.15g\n", p0);
    log_info(stderr, "Final rho value = %.15g\n", rho);
    log_info(stderr, "Total iterations of EM - %d\n", iter_counter-1);

    using namespace thrust::placeholders;
    thrust::transform(ez.begin(), ez.end(), idrLocal.begin(), (double) 1.0 - _1);
}
